body {
  margin: 0;
  overflow: hidden;
}

#viewport {
  position: absolute;
  top: 0;
  left: 0;
  right: 0;
  bottom: 0;
}

#view-buttons {
  position: absolute;
  bottom: 0;
  left: 0;
  padding: 1ex;
}

#control-buttons {
  position: absolute;
  bottom: 0;
  right: 0;
  padding: 1ex;
}

button {
  cursor: pointer;
}
