/******************************************************************************
 * Copyright © 2012-2014 Institut für Nachrichtentechnik, Universität Rostock *
 * Copyright © 2006-2012 Quality & Usability Lab,                             *
 *                       Telekom Innovation Laboratories, TU Berlin           *
 *                                                                            *
 * This file is part of the SoundScape Renderer (SSR).                        *
 *                                                                            *
 * The SSR is free software:  you can redistribute it and/or modify it  under *
 * the terms of the  GNU  General  Public  License  as published by the  Free *
 * Software Foundation, either version 3 of the License,  or (at your option) *
 * any later version.                                                         *
 *                                                                            *
 * The SSR is distributed in the hope that it will be useful, but WITHOUT ANY *
 * WARRANTY;  without even the implied warranty of MERCHANTABILITY or FITNESS *
 * FOR A PARTICULAR PURPOSE.                                                  *
 * See the GNU General Public License for more details.                       *
 *                                                                            *
 * You should  have received a copy  of the GNU General Public License  along *
 * with this program.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                            *
 * The SSR is a tool  for  real-time  spatial audio reproduction  providing a *
 * variety of rendering algorithms.                                           *
 *                                                                            *
 * http://spatialaudio.net/ssr                           ssr@spatialaudio.net *
 ******************************************************************************/

/// @file
/// QGUIFrame

#ifndef SSR_QGUIFRAME_H
#define SSR_QGUIFRAME_H

#include <QtCore/QTimer>
#include <QtGui/QMouseEvent>
#include <QtWidgets/QLabel>
#include <QtWidgets/QWidget>

//#include "qclicktextlabel.h"

/// QGUIFrame
class QGUIFrame : public QWidget
{
  Q_OBJECT

  public:
    QGUIFrame( QWidget* parent = 0, int top = 0, int bottom = 0, int left = 0,
        int right = 0);
    ~QGUIFrame();

    void resize(const int top, const int bottom,
		const int left, const int right);
    void set_text(const QString text);

  private:
    // Components of frame
    QLabel* _top;
    //QClickTextLabel* _bottom;
    QLabel* _bottom;
    QLabel* _left;
    QLabel* _right;

    QTimer* _clear_text_timer;

  private slots:
    void _clear_text();

  protected:
    virtual void mousePressEvent(QMouseEvent *event);
    virtual void mouseMoveEvent(QMouseEvent *event);
};

#endif
